#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This script extracts the ASN1 definition and JSON definitions from a rawx text file.
# and generates files that contains the extraction
#
# To obtain the raw file from a word document : open it
# Then in "view" menu, select normal, draft or web layout (any kind that removes page header and footers)
# Finally save the document as a text file
# launch the scrypt indicating rw files to be processed or directory owning the raw files.
#
# IDEMIA
# By Elder Dos Santos
# 2018 Elder Dos Santos
#

""" This module aims to extract ASN1 and JSON from raw text files """

import os           # for filesystem access
import sys          # for arguments
import re           # for regular expressions
import argparse     # parsing command line arguments

# script version
ExtractorVersion = "0.8.0"

# ASN1 filter data
ASN1_NAME = 'ASN1'
ASN1_START = r'-- ASN1START'
ASN1_STOP = r'-- ASN1STOP'
ASN1_EXTENTION = ".asn1"
# JSON filter data
JSON_NAME = 'JSON'
JSON_START = r'-- JSONSTART'
JSON_STOP = r'-- JSONSTOP'
JSON_EXTENTION = ".json"

# Specific for ASN1 with C style comments
COMMENT_START = r'^.*\/\*'
COMMENT_STOP = r'^.*\*\/.*$'
C_COMMENT_START = r'/*'
C_COMMENT_STOP = r'*/'
ASN1_COMMENT = r'-- '

# command line argument
F_ALL = 'ALL'

# general input extension
TXT_EXTENTION = ".txt"
# define for regex to check that the tag is at start of the line
RE_START = r'^.*'
RE_STOP = r'.*$'


class TagFilter(object):
    """Simple class to describe a filter and basic operation on a line for that filer"""
    def __init__(self, name, start, stop, ext):
        """class initialisation"""
        self.name = name
        self.start = start
        self.stop = stop
        self.ext = ext
        self.toggle = False
        # just keep in mind that match start from the beginning if you just search for a pattern use search
        self.matchstart = re.compile(RE_START + self.start + RE_STOP)
        self.matchstop = re.compile(RE_START + self.stop + RE_STOP)

    def init_filter_toggle(self):
        """re initialise the toggle"""
        self.toggle = False

    def line_to_keep(self, line, infilename, inum):
        """Ask for the provided line if this line should be keept accoridng the filter and it's state"""
        # regexp
        matchstart = self.matchstart.match(line)
        matchstop = self.matchstop.match(line)

        # if the line does not match the filters
        if matchstart is None and matchstop is None:
            # keep the line if toggle is set
            if self.toggle:
                return line
        else:
            # current line in string: an error might happen
            curline = str(inum + 1)

            # consistency check unapplicable for the moment
            if matchstart is not None and matchstop is not None:
                print("In file '" + infilename + "', in line :" + curline)
                print("\tboth :'" + self.start + "' and '" + self.start + "' detected.")

            # toggle to start
            if matchstart:
                # consistency check
                if self.toggle:
                    print("In file '" + infilename + "', in line :" + curline)
                    print("\tSecond '" + self.start + "' detected, seems that a '" + self.stop + "' is missing")
                # toggle : from now keep the lines
                self.toggle = True

            # toggle to stop
            if matchstop:
                # consistency check
                if not self.toggle:
                    print("In file '" + infilename + "', in line :" + curline)
                    print("\tSecond '" + self.stop + "' detected, seems that a '" + self.start + "' is missing")
                # toggle : from now don't keep the lines
                self.toggle = False
        # by default alway drop the line
        return ''


class TagASN1Filter(TagFilter):
    """Specific filter class for ASN1, with C style comment filtering/replacement"""
    def __init__(self):
        """class initialisation"""
        super().__init__(ASN1_NAME, ASN1_START, ASN1_STOP, ASN1_EXTENTION)
        # specific for ASN1
        self.comment_start = re.compile(COMMENT_START)
        self.comment_stop = re.compile(COMMENT_STOP)
        self.comment_toggle = False

    def init_filter_toggle(self):
        """re initialise the toggle for ASN1 comments"""
        # call the base class
        super().init_filter_toggle()
        self.comment_toggle = False

    def line_to_keep(self, line, infilename, inum):
        """Overriding of the root base class, only for ASN1 to handle comments"""
        # call the base class for filtering
        line = super().line_to_keep(line, infilename, inum)
        # we do something, but only when we are processing ASN1...
        if len(line) != 0:
            # regexp check for comment
            match_comment_start = self.comment_start.match(line)
            match_comment_stop = self.comment_stop.match(line)
            # if one line comment
            if match_comment_start and match_comment_stop:
                # change the input line !!!! put comment at the begin and remove the end. no toggle
                line = line.replace(C_COMMENT_START, ASN1_COMMENT)
                line = line.replace(C_COMMENT_STOP, '')
            else:
                # during comments (but not single line comments)
                if self.comment_toggle and not match_comment_start and not match_comment_stop:
                    # change the input line !!!!
                    line = ASN1_COMMENT + line
                # start a comment
                if match_comment_start:
                    # toggle current input
                    self.comment_toggle = True
                    # change the input line !!!! replace comment
                    line = line.replace(C_COMMENT_START, ASN1_COMMENT)
                # end a comment
                if match_comment_stop:
                    # toggle current input
                    self.comment_toggle = False
                    # change the input line !!!! remove comment (already done as toggle was on)
                    line = ASN1_COMMENT + line
                    line = line.replace(C_COMMENT_STOP, '')
        # in any case we keep the line
        return line


class JobTask(object):
    """This contains necessary the data to process a file"""
    def __init__(self, infiles, filterlist, outdir=None):
        """Initialise the members"""
        self.infilelist = infiles
        self.filterlist = filterlist
        self.outdir = outdir
        # internal values
        self.outfiles = list()      # list of output list files! (on list per input)
        # internal job of creating the output files !
        self.setoutputfiles()

    def setoutputfiles(self):
        """Defines the output files names for all input files. Done once to check easily output names collisions."""

        # intermediate list, just for check
        out_list = list()
        # create the list for all files
        for infile in self.infilelist:
            # get path and filename from input
            outpath, file_name = os.path.split(infile)
            # define an output file list as many filters...
            outputfilelist = list()
            # for all filters
            for afilter in self.filterlist:
                if self.outdir is not None:
                    # set ouput path to the required directory
                    outpath = self.outdir
                # compose output file fr this filter
                fileout = os.path.join(outpath, file_name + afilter.ext)
                # check if a file has exactly the same name.
                if out_list.count(fileout) != 0:
                    # adds in the name the number of counted value plus one... let the user search and found
                    ren_fileout = os.path.join(outpath, file_name + '_' + str(out_list.count(fileout) + 1) + afilter.ext)
                    # TODO better display
                    # Display something
                    msg = "Warning: out file '{}' generated from '{}' is renamed to '{}', "
                    msg += "because in the output directory '{}' this file will already be generated"
                    print(msg.format(fileout, infile, ren_fileout, outpath))
                    fileout = ren_fileout
                # ok now keep it in the list for this input file
                outputfilelist.append(fileout)
                # and in the global list
                out_list.append(fileout)
            # done for this input file, keep the ouput list
            self.outfiles.append(outputfilelist)

    def process(self):
        """Process the filters for all files"""
        # loop on all input files
        for idx, infile in enumerate(self.infilelist):
            print("--- Processing extraction for '" + infile + "'")
            # creation of a list of output handlers
            outhdlrlist = list()
            # open a file for all outputs needed. outfiles is a list of list...
            for outfile in self.outfiles[idx]:
                # open and output file and keep it in an ouput file list
                hdlr = open(outfile, 'w', encoding='utf-8')
                outhdlrlist.append(hdlr)
            # open input file
            inhdlr = open(infile, 'r')
            # process the file
            self.filefilter(idx, inhdlr, outhdlrlist)
            # closing...
            inhdlr.close()
            for hdlr in outhdlrlist:
                hdlr.close()

    def filefilter(self, idx, inhdlr, outhdlrlist):
        """This applies filter(s) to all lines of the file provided"""
        curline = "undefined"
        # restart the filters
        for flt in self.filterlist:
            flt.init_filter_toggle()
        try:
            # all lines of input file
            for inum, inline in enumerate(inhdlr):
                # for all filters
                for filt_idx, flt in enumerate(self.filterlist):
                    # exam input line
                    line = inline
                    # is line to be kept ?
                    line = flt.line_to_keep(line, self.infilelist[idx], inum)
                    # keep the line if toggle is set
                    if len(line) != 0:
                        outhdlrlist[filt_idx].write(line)
                # end of loop on filter, go to next line
            # processing finished on all file
        except TypeError as error:
            print("Something wrong ! stopped in file:'" + self.infilelist[idx] + "' at line " + curline)
            raise error


def getargs():
    """Get arguments from command line"""

    # general dewcription
    desc_summary = "Extract snipplet(s) of code from files, version {}".format(ExtractorVersion)
    # post description
    longhelp = "Extract {} snipplet(s) of code delimited by '{}' and '{}' or ".format(ASN1_NAME, ASN1_START, ASN1_STOP)
    longhelp += "{} snipplet(s) delimited by '{}' and '{}'. Usage:".format(JSON_NAME, JSON_START, JSON_STOP)
    longhelp += __file__ + " <input_file_1>... <directory_1>...\n"
    longhelp += "The option -e|--ext [ext1 ext2] allow to select files with the given extension(s) in the directories,\n"
    longhelp += "\tIf not provided all files with '{}' of directory will be processed.\n".format(TXT_EXTENTION)
    longhelp += "The option -o|--out [<directory>] define where output(s) snipplet(s) will be stored.\n"
    longhelp += "\tIf missing, the directory of the input file will be used.\n"
    longhelp += "The option -f|--filter FILTER allow to select the type of snipplet to extract.\n"
    longhelp += "\tIf not provided all sniplets are retrieved.\n"

    # global description
    # parser = argparse.ArgumentParser(description=desc_summary, epilog=longhelp)
    parser = argparse.ArgumentParser(description=desc_summary)
    # postional argument : inputs files to be processed, can be empty in that case current directory is used
    # TODO add recursivity (-r) for directories inside)
    helper = "List of files to process or directories where files to process will be found. "
    helper += "Search for files in directories is not recursive."
    parser.add_argument('input', help=helper, nargs='*', default=".")
    # actions to do (filter required)
    helper = "Type of code snipplet to be extracted, default: {}".format(F_ALL)
    parser.add_argument("-f", "--filter", help=helper, choices=[ASN1_NAME, JSON_NAME, F_ALL], default=F_ALL)
    # list of extension for directories, shall not be empty
    helper = "List of extension of files to be selected in directories given in input, "
    helper += "default: '{}'. Should start with a '.'".format(TXT_EXTENTION)
    parser.add_argument('-e', '--ext', help=helper, nargs='+', default=[TXT_EXTENTION])
    # output directory
    helper = "Output directory: if specified, all outputs files will be placed there, default: <current directory>."
    parser.add_argument('-o', '--out', help=helper, default=None)
    # parse input
    theargs = parser.parse_args()

    # return the nice structure
    return theargs


if __name__ == '__main__':
    """ main input """

    # call argument parser
    args = getargs()

    # check python version
    if sys.version_info < (3, 7):
        print("Caution: tested with python 3.7 ...")

    # create list of files to process
    # TODO recursivity here might be needed
    files_to_process = list()
    # if a filename is given
    for IN in args.input:
        # if it's a file, add it in a list
        if os.path.isfile(IN):
            # create list for extraction with a single file
            files_to_process.append(IN)
        # input is a directory
        elif os.path.isdir(IN):
            # get a list of all files to be processed (i had help from https://stackoverflow.com/questions/2225564/get-a-filtered-list-of-files-in-a-directory)
            files_to_process += [os.path.join(IN, x) for x in os.listdir(IN) if any(x.endswith(ext) for ext in args.ext)]
        else:
            print("Unable to process input '{}' not recognised as a file nor a directory, exiting...".format(IN))
            exit()

    # check that output exist
    if args.out is not None and not os.path.isdir(args.out):
        print("Specified ouput '{}' is not a directory or don't exist. Process aborted".format(args.out))
        exit()

    # ASN1 filter
    ASN1Filter = TagASN1Filter()
    # JSON Filter
    JSONFilter = TagFilter(JSON_NAME, JSON_START, JSON_STOP, JSON_EXTENTION)
    # create the list of filters to apply
    # TODO we can do better to have something more generic there (list of filter and the associated object)
    filters = list()
    if args.filter == ASN1_NAME or args.filter == F_ALL:
        # add ASN1 filter
        filters.append(ASN1Filter)
    if args.filter == JSON_NAME or args.filter == F_ALL:
        # add ASN1 filter
        filters.append(JSONFilter)

    # process all files
    if len(files_to_process) > 0:
        # create a job with all files and filters and process them
        job = JobTask(files_to_process, filters, args.out)
        job.process()
    else:
        print("Nothing processed, empty directory or no files matching the provided extention")
